/* 
 *   IWSCANSWITCH
 *       CEPSR, Columbia University
 *
 *       Rajesh Menon - CIMS, NYU - 12/31/2003
 *       prm225@cs.nyu.edu
 * 
 * Tool for the 7DS project at Columbia University to scan
 * for the availability of infrastructure access-points and 
 * switch to 7DS if needed, and vice-versa.
 * 
 * Parts from: iwconfig.c, iwlist.c, wireless.h, iwlib.h
 * Acknowledgments: Jean II [HPL]
 *
 * Link with: iwlib.c and lm
 *
 */

#include <iwlib.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define IFNAMESIZ 50
#define ADHOC "Ad-Hoc"
#define ADHOCESSID "7DS"

// 2.5 and 1 sec delay for usleep()
#define SLEEP 2500000  
#define MINISLEEP 1000000

// Number of scans allowed with low SNR
#define MAXNUMSCANS 3

// SNR+GRACE will switch immediately without checks for averages
#define GRACE 10

/* Err messages on set_info (might as well)
 * If not wanted, call the iw_get/set_ext() functions direct */
#define IW_GET_EXT_ERR(skfd, ifname, request, wrq, rname) \
	do { \
	if(iw_get_ext(skfd, ifname, request, wrq) < 0) { \
		ERR_SET_EXT(rname, request); \
		fprintf(stderr, "    GET failed on device %-1.16s ; %s.\n", \
			ifname, strerror(errno)); \
		return(-6); \
	} } while(0)

#define IW_SET_EXT_ERR(skfd, ifname, request, wrq, rname) \
	do { \
	if(iw_set_ext(skfd, ifname, request, wrq) < 0) { \
		ERR_SET_EXT(rname, request); \
		fprintf(stderr, "    SET failed on device %-1.16s ; %s.\n", \
			ifname, strerror(errno)); \
		return(-5); \
	} } while(0)

#define ERR_SET_EXT(rname, request) \
	fprintf(stderr, "Error for wireless request \"%s\" (%X) :\n", \
		rname, request)

	
struct iwconfig {
  // only need essid, mac and mode here
  char name[IFNAMESIZ + 1];
  int essid_on;
  int has_essid;
  char essid[IW_ESSID_MAX_SIZE + 1];
  int has_ap_addr;
  sockaddr ap_addr;
  int has_mode;
  int mode;
  iwstats stats;
  int has_stats;  
  int slevel;
  int nlevel;
};


struct iwlist {
  // only what is needed to be decoded
  int srnum;
  char mac[18];
  char essid[IW_ESSID_MAX_SIZE+1];
  char mode[25];
  int slevel;
  int nlevel;
  unsigned char key[4];
  struct iwlist *next;
};


int get_iwconfig_info (int skfd, char *ifname, struct iwconfig *info);
int get_iwlist_info (int skfd, char *ifname, struct iwlist *);

// decode events and populate the linklist of APs
int get_iwlist_tokens (
  struct iw_event *event, 
  int ap_num, 
  struct iw_range *iwrange, 
  int has_range, 
  struct iwlist *list); 

// MODE: 1=ad-hoc, 2=client (managed)
int switchto (int skfd, char *ifname, int mode, char essid[IW_ESSID_MAX_SIZE + 1]);

void display_iwconfig_info (struct iwconfig *info, char *ifname);
void display_iwlist_info (struct iwlist *list);

// Housekeeping
void initval (int *lowsnrcount, int *snrsum, int *snravg);
void deletelist (struct iwlist *list);

void usage (char pgm_name[]);


int 
main (int argc, char *argv[]) 
{
  // generic raw socket desc.
  int skfd;
  
  int snr;
  int snrsum = 0;
  int snravg;
  int count = 0;
  int ret;
  int snrtmp;
  
  char *dev;

  // iwconfig stat holder
  struct iwconfig *piwconfig;
  piwconfig = (struct iwconfig *) malloc (sizeof (struct iwconfig));

  // linklist for iwlist tokens
  struct iwlist *list;
  list = (struct iwlist *) malloc (sizeof (struct iwlist));
  list->next = NULL;
 
  // create a channel to the NET kernel
  if ((skfd = iw_sockets_open ()) < 0) {
    perror ("socket");
    exit (-1);
  }
 
  if (argc != 3)
    usage (argv[0]);

  dev = argv[1];
  snr = (int) atoi (argv[2]);
 
  // forever
  while (1) {
    system ("clear");	  
    printf ("SNR Threshold = %d\n", snr);
    if (count > 0)
      printf ("Scan count = %d\n", count);

    // get iwconfig output here
    get_iwconfig_info (skfd, dev, piwconfig);
    display_iwconfig_info (piwconfig, dev);
 
    /* Start checks here using info in 'piwconfig' and 'list'  */
    /* In Client mode */
    if (strncasecmp (iw_operation_mode[piwconfig->mode], ADHOC, strlen (ADHOC))) {
      // Get scan results
      get_iwlist_info (skfd, dev, list);
      
      if (list != NULL) {
        /* SNR is below the threshold */
        display_iwlist_info (list);
        snrtmp = piwconfig->slevel - piwconfig->nlevel;
        if (snrtmp < snr) {
          printf ("SNR falling below %d \t Scanning %d...\n", snr, count);

          /* Num of scans exceeds limit */
          if (count > MAXNUMSCANS) {
            // Switch to 7DS only if Avg.SNR collected is less than threshold value
            if (snravg < snr) {
              printf ("Switch to 7ds\n");
              switchto (skfd, dev, 1, ADHOCESSID);
	      initval (&count, &snrsum, &snravg);
	    }
	    else 
              initval (&count, &snrsum, &snravg);	
	  }	  
	  else {
            count++;
	    snrsum += list->slevel - list->nlevel;
	    snravg = snrsum/count;	  
	  }
	}
	/* SNR is good */
	else 
          initval (&count, &snrsum, &snravg);
      }
      // list is NULL
      else {
        count++;
        snrsum += snr+1;
        snravg = snrsum/count;	  
      }
    } // finish checks in client mode
      
    /* In Ad-Hoc mode */
    else {
      /* Get iwlist output into list
       * Do a check on the current ESSID, and then perform the scan
       * This is because scanning while in ad-hoc works for essid 'any'
       */
      if (!strncasecmp (piwconfig->essid, ADHOCESSID, strlen (ADHOCESSID))) {
        switchto (skfd, dev, 1, "any");  
        ret = -2;
      }
      get_iwlist_info (skfd, dev, list);
      switchto (skfd, dev, 1, ADHOCESSID);
     
      if ((list != NULL) && (!strncasecmp (list->key, "off", strlen ("off")))) {
        display_iwlist_info (list);
        snrtmp = list->slevel - list->nlevel;
        if (snrtmp >= (snr + GRACE)) {
          printf ("AP with excellent SNR seen. \nSwitching now.\n");
	  switchto (skfd, dev, 2, list->essid);
	  system ("dhclient eth0");
	  initval (&count, &snrsum, &snravg);
	}
	else {
          if (count > MAXNUMSCANS) {
            if (snravg > snr) {
              printf ("AP with good SNR seen. \nSwitching now.\n");
	      switchto (skfd, dev, 2, list->essid);
	      system ("dhclient eth0");
	      initval (&count, &snrsum, &snravg);  
	    }
            else
              initval (&count, &snrsum, &snravg);  
	  }
	  // Revert back to ADHOCESSID
	  else {
            switchto (skfd, dev, 1, ADHOCESSID);
	    count++;
	    snrsum += snrtmp;
	    snravg = snrsum/count;	  
	  }
	}
      }
    }
    /* end of checks */
   
    // cleanup
    deletelist (list);
    usleep (SLEEP);  
  }

  // cleanup 
  close (skfd);
  return 1;
}  // main()


int
get_iwconfig_info(int skfd, char *ifname, struct iwconfig *info)
{
  struct iwreq wrq;
  memset((char *) info, 0, sizeof(struct wireless_info));

  /* Get wireless name */
  if(iw_get_ext(skfd, ifname, SIOCGIWNAME, &wrq) < 0) {
    /* If no wireless name : no wireless extensions */
    /* But let's check if the interface exists at all */
    struct ifreq ifr;

    strcpy(ifr.ifr_name, ifname);
    if(ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0)
      return(-ENODEV);
    else
      return(-ENOTSUP);
  }
  else {
    strncpy(info->name, wrq.u.name, IFNAMSIZ);
    info->name[IFNAMSIZ] = '\0';
  }

  /* Get ESSID */
  wrq.u.essid.pointer = (caddr_t) info->essid;
  wrq.u.essid.length = IW_ESSID_MAX_SIZE + 1;
  wrq.u.essid.flags = 0;
  if(iw_get_ext(skfd, ifname, SIOCGIWESSID, &wrq) >= 0) {
    info->has_essid = 1;
    info->essid_on = wrq.u.data.flags;
  }

  /* Get AP address */
  if(iw_get_ext(skfd, ifname, SIOCGIWAP, &wrq) >= 0) {
    info->has_ap_addr = 1;
    memcpy(&(info->ap_addr), &(wrq.u.ap_addr), sizeof (sockaddr));
  }

  /* Get operation mode */
  if(iw_get_ext(skfd, ifname, SIOCGIWMODE, &wrq) >= 0) {
    info->mode = wrq.u.mode;
    if((info->mode < IW_NUM_OPER_MODE) && (info->mode >= 0))
      info->has_mode = 1;
  }

  /* Get signal stats */
  if(iw_get_stats(skfd, ifname, &(info->stats)) >= 0) {
      info->has_stats = 1;
  }
  if (info->has_stats == 1) {
    info->slevel = info->stats.qual.level - 0x100;
    info->nlevel = info->stats.qual.noise - 0x100;
  }

  return(0);
}  // get_iwconfig_info()


int
get_iwlist_tokens (
  struct iw_event *event, 
  int ap_num, 
  struct iw_range *iwrange, 
  int has_range, 
  struct iwlist *node)
{
  /* Now, let's decode the event */
  switch(event->cmd) {
    case SIOCGIWAP:
      {
        node->srnum = ap_num-1;
        iw_pr_ether(node->mac, event->u.ap_addr.sa_data);
      }
      break;
    case SIOCGIWMODE:
      {
	strncpy (node->mode, iw_operation_mode[event->u.mode], 14);
        node->mode[15] = '\0';
      }  
      break;
    case SIOCGIWESSID:
      {
	if((event->u.essid.pointer) && (event->u.essid.length))
	  strncpy(node->essid, event->u.essid.pointer, event->u.essid.length);
	node->essid[event->u.essid.length] = '\0';
      } 
      break;
    case IWEVQUAL:
      {
	node->slevel = event->u.qual.level - 0x100;
	node->nlevel = event->u.qual.noise - 0x100;
      }	
      break;
    case SIOCGIWENCODE:
      {
	if(event->u.data.flags & IW_ENCODE_DISABLED) 
	  strncpy (node->key, "off", sizeof ("off"));
	else 
	  strncpy (node->key, "on", sizeof ("on"));
	ap_num++;
      }
      break;
    default:
      break; 
  }	/* switch(event->cmd) */

  /* May have changed */
  return(ap_num);
} // get_iwlist_tokens()


int
get_iwlist_info(int skfd, char *ifname, struct iwlist *list)
{
  struct iwreq wrq;
  unsigned char buffer[IW_SCAN_MAX_DATA];	/* Results */
  struct timeval tv;				/* Select timeout */
  int timeout = 5000000;		/* 5s */
  struct iwlist *tmp;
  
  tmp = list;

  /* Init timeout value -> 250ms*/
  tv.tv_sec = 0;
  tv.tv_usec = 250000;

  /* set the IW_SCAN_ flags properly */
  wrq.u.param.flags = IW_SCAN_DEFAULT;
  wrq.u.param.value = 0;		/* Later */

  /* Initiate Scanning */
  if(iw_set_ext(skfd, ifname, SIOCSIWSCAN, &wrq) < 0) {
    if(errno != EPERM) {
      fprintf(stderr, "%-8.8s  Interface doesn't support scanning : %s\n\n", ifname, strerror(errno));
      return(-1);
    }
    tv.tv_usec = 0;  
  }
  timeout -= tv.tv_usec;

  /* Forever */
  while(1) {
    fd_set rfds;		/* File descriptors for select */
    int last_fd;	/* Last fd */
    int ret;

    /* Guess what ? We must re-generate rfds each time */
    FD_ZERO(&rfds);
    last_fd = -1;

    /* Wait until something happens */
    ret = select(last_fd + 1, &rfds, NULL, NULL, &tv);

    /* Check if there was an error */
    if(ret < 0) {
      if(errno == EAGAIN || errno == EINTR)
        continue;
      fprintf(stderr, "Unhandled signal - exiting...\n");
      return(-1);	
    }

    /* Check if there was a timeout */
    if(ret == 0) {
      /* Try to read the results */
      wrq.u.data.pointer = buffer;
      wrq.u.data.flags = 0;
      wrq.u.data.length = sizeof(buffer);
      if(iw_get_ext(skfd, ifname, SIOCGIWSCAN, &wrq) < 0) {
        /* Check if results not available yet */
        if(errno == EAGAIN) {
          /* Restart timer for only 100ms*/
	  tv.tv_sec = 0;
	  tv.tv_usec = 100000;
	  timeout -= tv.tv_usec;
	  if(timeout > 0)
	    continue;	/* Try again later */
	}

        /* Bad error */
        fprintf(stderr, "%-8.8s  Failed to read scan data : %s\n\n", ifname, strerror(errno));
        return(-2); 
	}
      else
        /* We have the results, go to process them */
        break; 
    } 
  }

  if(wrq.u.data.length) {
    struct iw_event iwe;
    struct stream_descr	stream;
    int ap_num = 1;
    int ret;
    struct iw_range range;
    int has_range;
    int check;
    int flag = 0;
    
    check = ap_num;
    iw_init_event_stream(&stream, buffer, wrq.u.data.length);
    do {
      /* Extract an event and print it */
      ret = iw_extract_event_stream (&stream, &iwe);
      if (ret > 0) { 
        if ((ap_num != 1) && (flag == 1)) { // Mem alloc only for new cells found
          tmp->next = (struct iwlist *) malloc (sizeof (struct iwlist));
	  tmp = tmp->next;
	  tmp->next = NULL;
	  flag = 0;
	  check = ap_num;
	}

        ap_num = get_iwlist_tokens (&iwe, ap_num, &range, has_range, tmp);	
        if (check != ap_num)
	  flag = 1;
      }
    } while(ret > 0);
  }
  else
    printf("%-8.8s  No scan results\n", ifname);

  return(0);
}  // get_iwlist_info()


int
switchto (int skfd, char *ifname, int mode, char essid[IW_ESSID_MAX_SIZE + 1])
{
  struct iwreq wrq;
  char tmpessid[IW_ESSID_MAX_SIZE + 1];
  memcpy (tmpessid, essid, IW_ESSID_MAX_SIZE+1);
  
  /* Change ESSID 
   * The flag settings are done differently for ad-hoc and client modes,
   * as otherwise scan results with APs having different names are 
   * shown only on ESSID promiscuous.
   */
  if (mode == 2) {
    if (!strncasecmp (essid, "any", strlen ("any"))) { 
      wrq.u.essid.flags = 1;
      tmpessid[0] = '\0';
    }
    else 
      wrq.u.essid.flags = 0;
  }
  else {
    if (!strncasecmp (essid, "any", strlen ("any"))) { 
      wrq.u.essid.flags = 0;
      tmpessid[0] = '\0';
    }
    else 
      wrq.u.essid.flags = 1;
  }
  
  wrq.u.essid.pointer = (caddr_t) tmpessid;
  wrq.u.essid.length = strlen(tmpessid) + 1;
  IW_SET_EXT_ERR(skfd, ifname, SIOCSIWESSID, &wrq, "Set ESSID");

  /* Change mode 
   * 1=adhoc, 2=client */
  wrq.u.mode = mode;
  IW_SET_EXT_ERR(skfd, ifname, SIOCSIWMODE, &wrq, "Set Mode");
  
  usleep (SLEEP);

  return(0);
} // switchto()


void initval (int *a, int *b, int *c) 
{
  *a = 0;
  *b = 0;
  *c = 0;
} // initval()


void
display_iwconfig_info(struct iwconfig *info, char *ifname)
{
  char buffer[128];	/* Temporary buffer */

  /* One token is more of less 5 characters, 14 tokens per line */
  int tokens = 3;	/* For name */

  /* Display device name and wireless name (name of the protocol used) */
  printf("%-8.8s  %s  ", ifname, info->name);

  /* Display ESSID (extended network), if any */
  if(info->has_essid) {
    if(info->essid_on) {
      /* Does it have an ESSID index ? */
      if((info->essid_on & IW_ENCODE_INDEX) > 1)
        printf("ESSID:\"%s\" [%d]  ", info->essid, (info->essid_on & IW_ENCODE_INDEX));
      else
        printf("ESSID:\"%s\"  ", info->essid);
    }
    else
      printf("ESSID:off/any  ");
  }

  /* Display the current mode of operation */
  if(info->has_mode) {
    printf("Mode:%s  ", iw_operation_mode[info->mode]);
    tokens +=3;
  }

  /* Display the address of the current Access Point */
  if(info->has_ap_addr) {
    /* A bit of clever formatting */
    if(tokens > 8) {
      printf("\n          ");
      tokens = 0;
    }

    tokens +=6;

  /* Oups(Not my typo, this!) ! No Access Point in Ad-Hoc mode */
  if((info->has_mode) && (info->mode == IW_MODE_ADHOC))
    printf("Cell:");
  else
    printf("Access Point:");
  printf(" %s  ", iw_pr_ether(buffer, info->ap_addr.sa_data));
  }

  printf("\n");
}  // display_iwconfig_info()


void display_iwlist_info (struct iwlist *list)
{
  struct iwlist *p;
  printf ("\nNR\tMAC Addr\t\tMode\tSignal\tNoise\tSNR\tEncr\tESSID \n");
  for ( p = list; p->next != NULL; p = p->next ) 
    printf ("%2d\t%s\t%s\t%d\t%d\t%d\t%s\t%s\n", 
      p->srnum+1, 
      p->mac, 
      p->mode, 
      p->slevel, 
      p->nlevel, 
      p->slevel - p->nlevel,  
      p->key, 
      p->essid);
  if (p->srnum != 0)
    printf ("%2d\t%s\t%s\t%d\t%d\t%d\t%s\t%s\n", 
      p->srnum+1, 
      p->mac, 
      p->mode, 
      p->slevel, 
      p->nlevel, 
      p->slevel - p->nlevel,  
      p->key, 
      p->essid);
} // display_iwlist_info()


void deletelist (struct iwlist *list)
{
  struct iwlist *p, *next;
  for ( p = list->next; p != NULL; p = next) {
    next = p->next;
    free (p);
  }
} // deletelist()


void usage (char pgm_name[])
{
  printf ("Usage: %s <interface> <SNR Threshold>\n", pgm_name);
  exit(-1);
} 


